from functools import wraps


def group_format(fun):
    @wraps(fun)
    def _group_format_wrapper(high_data, *args, **kwargs):
        if not isinstance(high_data, (list, tuple)):
            high_data = [high_data]
        return fun(high_data, *args, **kwargs)

    return _group_format_wrapper



from json import dumps
from json import loads


NONE_TYPE = type(None)
STATE_INFO = {"fun", "order", "state"}


def dump_key(key):
    return dumps(key)


def load_key(key):
    return tuple(loads(key))


def path_ref(data, key):
    for p in key[:-1]:
        data = data[p]
    return data


def del_path(data, key):
    # key cant be empty
    assert key
    for p in key[:-1]:
        data = data[p]
    del data[key[-1]]


def path_data(data, key):
    for p in key:
        data = data[p]
    return data


def path_replace(data, key, value):
    # key cant be empty
    assert key
    for p in key[:-1]:
        data = data[p]
    data[key[-1]] = value


def paths(data, include_dunder=False, include_state_info=False, at=None):
    if at is None:
        at = []
    if isinstance(data, (bool, int, float, str, NONE_TYPE)):
        return
    elif isinstance(data, list):
        for index, item in enumerate(data):
            new_at = at.copy()
            new_at.append(index)
            yield tuple(new_at)
            yield from paths(item, include_dunder, include_state_info, new_at)
        return
    elif isinstance(data, dict):
        for key, item in data.items():
            if (
                not include_dunder
                and isinstance(key, str)
                and key.startswith("__")
                and key.endswith("__")
            ):
                continue
            if not include_state_info and key in STATE_INFO:
                continue
            new_at = at.copy()
            new_at.append(key)
            yield tuple(new_at)
            yield from paths(item, include_dunder, include_state_info, new_at)
        return
    raise NotImplementedError()



from copy import deepcopy
from pprint import pprint

from idem_data_insights.idem_data_insights import high
from idem_data_insights.idem_data_insights import low
from idem_data_insights.idem_data_insights import rpath
from idem_data_insights.idem_data_insights.group_format import group_format


@group_format
def state_count(high_data):
    count_groups = []
    for group in high_data:
        count_group = []
        for state, count in group[high.STATE_COUNT].items():
            count_group.append((state, count))
        count_groups.append(
            sorted(count_group, key=lambda k: (k[1], k[0]), reverse=True)
        )
    return count_groups


def state_count_pp(data):
    for index, group in enumerate(data):
        print(f"#### {index}")
        for state, count in group:
            print(f"{state}, {count}")


def _variation_sort_fun(v):
    return (
        len(v["PATH"]),
        abs(50 - v[high.UNIQUE]),
        abs(0.5 - v[high.AVERAGE] / v[high.COUNT]),
        v[high.COUNT],
    )


@group_format
def variation_sort(high_data, sort_fun=_variation_sort_fun):
    variation_groups = []
    for group in high_data:
        variation_group = []
        for path, path_variation in group[high.PATH_VARIATIONS].items():
            variation_group.append({"PATH": path, **path_variation})
        variation_groups.append(variation_group)
    return [sorted(v_group, key=sort_fun) for v_group in variation_groups]


def variation_sort_pp(data):
    for index, group in enumerate(data):
        print(f"#### {index}")
        for v in group:
            print(
                f"{v['PATH']}, COUNT: {v[high.COUNT]}, UNIQUE: {v[high.UNIQUE]}, AVERAGE: {v[high.AVERAGE]}"
            )


@group_format
def mutable_data(high_data, reverse=False):
    m_data = []
    for group in high_data:
        new_group = []
        group = deepcopy(group)
        for low_state in group[high.STATES]:
            # sorting and reverse avoids crash because of list
            paths = list(low_state[low.INSIGHTS_LOW][low.PATHS])
            paths.sort(reverse=True)
            for path in paths:
                if reverse:
                    if path in group[high.MUTABLE] and path in group[high.CHILDREN]:
                        rpath.del_path(low_state, path)
                else:
                    if path in group[high.IMMUTABLE]:
                        rpath.del_path(low_state, path)
            # remove all dunders
            for key in list(low_state.keys()):
                if isinstance(key, str) and key.startswith("__") and key.endswith("__"):
                    del low_state[key]
                elif key in {"fun", "order", "state"}:
                    del low_state[key]
            new_group.append(low_state)
        m_data.append(new_group)
    return m_data


def mutable_data_pp(data):
    for index, group in enumerate(data):
        print(f"#### {index}")
        for g in group:
            pprint(g)



from functools import partial

from idem_data_insights.idem_data_insights import build
from idem_data_insights.idem_data_insights import high
from idem_data_insights.idem_data_insights.group_format import group_format


def null_comp(high_data, other_high_data):
    return False


def state_fun_comp(high_data, other_high_data):
    for state in high_data[high.STATE_COUNT]:
        if state in other_high_data[high.STATE_COUNT]:
            return True
    return False


def build_path_comp(path):
    return partial(path_comp, path=path)


def path_comp(high_data, other_high_data, path):
    if path not in high_data[high.PATHS_DATA]:
        return False
    if path not in other_high_data[high.PATHS_DATA]:
        return False
    return bool(
        high_data[high.PATHS_DATA][path].intersection(
            other_high_data[high.PATHS_DATA][path]
        )
    )


@group_format
def split(high_data, comp_fun):
    new_high = []
    for h in high_data:
        low_split = [[s] for s in h[high.STATES]]
        high_split = build.rebuild_sls(low_split)
        high_joins = {i: {i} for i in range(len(high_split))}
        for index, hs in enumerate(high_split):
            for other_index, other_hs in enumerate(high_split[index + 1 :]):
                other_index += index + 1
                if high_joins[index] != high_joins[other_index] and comp_fun(
                    hs, other_hs
                ):
                    high_joins[index].update(high_joins[other_index])
                    high_joins[other_index].update(high_joins[index])
                    for i in high_joins[index]:
                        high_joins[i].update(high_joins[index])

        for i in range(len(high_joins)):
            if i not in high_joins:
                continue
            new_high.extend(
                build.rebuild_sls([h[high.STATES][s] for s in high_joins[i]])
            )
            for dead_join in high_joins[i]:
                del high_joins[dead_join]
    return new_high


# TODO merge
def merge(high_data, comp_fun):
    pass



def __init__(hub):
    # Remember not to start your app in the __init__ function
    # This function should just be used to set up the plugin subsystem
    # The run.py is where your app should usually start
    for dyne in ["idem"]:
        hub.pop.sub.add(dyne_name=dyne)


def cli(hub):
    hub.pop.config.load(["idem_data_insights"], cli="idem_data_insights")
    # Your app's options can now be found under hub.OPT.idem_data_insights
    kwargs = dict(hub.OPT.idem_data_insights)

    # Initialize the asyncio event loop
    hub.pop.loop.create()

    # Start the async code
    coroutine = hub.idem_data_insights.init.run(**kwargs)
    hub.pop.Loop.run_until_complete(coroutine)


async def run(hub, **kwargs):
    """
    This is the entrypoint for the async code in your project
    """
    # Get sls data
    sls_data = (await hub.idem_data_insights.compiler.compile())["low"]
    # get high sls data
    hub.idem_data_insights.build.build(sls_data)
    print("Done")



from idem_data_insights.idem_data_insights import okey
from idem_data_insights.idem_data_insights import rpath

INSIGHTS_LOW = "__idem_data_insights_low__"
PATHS = "PATHS"
PATHS_INVERSE = "PATHS_INVERSE"


def low(sls_data):
    for state in sls_data:
        state[INSIGHTS_LOW] = {PATHS: {}, PATHS_INVERSE: {}}
        for path in rpath.paths(state):
            data = okey.object_key(rpath.path_data(state, path))
            state[INSIGHTS_LOW][PATHS][path] = data
            if data not in state[INSIGHTS_LOW][PATHS_INVERSE]:
                state[INSIGHTS_LOW][PATHS_INVERSE][data] = set()
            state[INSIGHTS_LOW][PATHS_INVERSE][data].add(path)
    return sls_data



from idem_data_insights.idem_data_insights import high
from idem_data_insights.idem_data_insights import low
from idem_data_insights.idem_data_insights.group_format import group_format


def build(hub, sls_data):
    if len(sls_data) and not isinstance(sls_data[0], (tuple, list)):
        sls_data = [sls_data]
    return [high.high(low.low(s)) for s in sls_data]


@group_format
def rebuild(high_data):
    return build([h[high.STATES] for h in high_data])


@group_format
def rebuild_high(high_data):
    return [high.high(h[high.STATES]) for h in high_data]


def rebuild_sls(sls_data):
    if len(sls_data) and not isinstance(sls_data[0], (tuple, list)):
        sls_data = [sls_data]
    return [high.high(s) for s in sls_data]



import pathlib


def get_refs(hub):
    """
    Determine where the sls sources are
    """
    sls_sources = []
    slses = []
    sls = hub.OPT.idem_data_insights.sls
    path = pathlib.Path(sls)
    if path.is_file():
        ref = str(path.stem if path.suffix == ".sls" else path.name)
        slses.append(ref)
        implied = f"file://{path.parent}"
        if implied not in sls_sources:
            sls_sources.append(implied)
    else:
        slses.append(sls)

    return {"sls_sources": sls_sources, "sls": slses}


async def compile(hub) -> dict:
    """
    Execute the cli routine to validate states
    """
    name = "compile"
    src = hub.idem_data_insights.compiler.get_refs()
    await hub.idem.state.validate(
        name=name,
        sls_sources=src["sls_sources"],
        render="jinja|yaml",
        runtime="serial",
        subs=["states"],
        cache_dir="cachedir",
        sls=src["sls"],
        test=True,
        acct_file=None,
        acct_key=None,
        acct_profile=None,
    )

    errors = hub.idem.RUNS[name]["errors"]
    if errors:
        display = hub.output.nested.display(errors)
        print(display)
        # Return a non-zero error code
        return len(errors)

    ret = {
        "high": hub.idem.RUNS[name]["high"],
        "low": hub.idem.RUNS[name]["low"],
        "meta": hub.idem.RUNS[name]["meta"],
    }
    return ret



from copy import deepcopy

from idem_data_insights.idem_data_insights import high
from idem_data_insights.idem_data_insights import low
from idem_data_insights.idem_data_insights import rpath
from idem_data_insights.idem_data_insights.group_format import group_format


@group_format
def state_fun_template_group(high_data):
    template_groups = {}
    for group in high_data:
        group_name = [
            f"{state}X{count}" for state, count in group[high.STATE_COUNT].items()
        ]
        group_name.sort()
        group_name = "-".join(group_name)
        if group_name not in template_groups:
            template_groups[group_name] = []
        group_order = {}
        for state in group[high.STATES]:
            state_name = f"{state['state']}.{state['fun']}"
            if state_name not in group_order:
                group_order[state_name] = []
            group_order[state_name].append(state)
        for _, states in group_order.items():
            states.sort(key=lambda s: s["order"])
        template_groups[group_name].append(group_order)
    return template_groups


def build_templates(templates_groups):
    templates = {}
    for template in templates_groups:
        # find all immutable keys
        grouping = []
        for group in templates_groups[template]:
            count = -1
            for state_key in group:
                for state in group[state_key]:
                    count += 1
                    if len(grouping) == count:
                        grouping.append([])
                    grouping[count].append(state)
        # remove sub keys
        grouping = [high.high(states) for states in grouping]
        grouping_keys = []
        for high_data in grouping:
            mutable = high_data[high.MUTABLE].copy()
            for key in mutable.copy():
                for i in range(-1, len(key) * -1, -1):
                    try:
                        mutable.remove(key[:i])
                    except KeyError:
                        pass
            grouping_keys.append(mutable)
        # link keys
        link_keys = []
        for spot, group in enumerate(grouping_keys):
            for key in group:
                link = {"key": key, "optional": grouping[spot][high.OPTIONAL][key]}
                if not link["optional"]:
                    high_count = 0
                    key_data = None
                    for name, count in grouping[spot][high.PATH_VARIATIONS][key][
                        high.VARIATIONS
                    ].items():
                        if len(count) > high_count:
                            high_count = len(count)
                            key_data = rpath.path_data(
                                grouping[spot][high.STATES][list(count)[0]], key
                            )
                    if (
                        high_count
                        / grouping[spot][high.PATH_VARIATIONS][key][high.COUNT]
                        > 0.5
                    ):
                        link["default"] = key_data
                link_keys.append({spot: [link]})
        # TODO link keys
        static_data = []
        for _, group in templates_groups[template][0].items():
            static_data.extend(group)
        for link in link_keys:
            for static_state_id, static_paths in link.items():
                static_state = static_data[static_state_id]
                for path in static_paths:
                    if path["optional"]:
                        try:
                            rpath.del_path(static_state, path["key"])
                        except (IndexError, KeyError):
                            pass
                    if "default" in path:
                        rpath.path_replace(static_state, path["key"], path["default"])
        templates[template] = {"links": link_keys, "static": static_data}
    return templates


def fill_templates(templates, templates_groups):
    filled_templates = {}
    for group_name, groups in templates_groups.items():
        if group_name not in filled_templates:
            filled_templates[group_name] = []
        for group in groups:
            spot = -1
            template_copy = deepcopy(templates[group_name])
            filled_templates[group_name].append(template_copy)
            del template_copy["static"]
            for state, states in group.items():
                for s in states:
                    spot += 1
                    for link in template_copy["links"]:
                        if spot in link:
                            try:
                                link["data"] = rpath.path_data(s, link[spot][0]["key"])
                            except (KeyError, IndexError):
                                pass
    return filled_templates


def unroll_templates(templates, filled_templates):
    states = []
    for filled_name, filled_template in filled_templates.items():
        template = templates[filled_name]
        for filled in filled_template:
            state = deepcopy(template["static"])
            for link in filled["links"]:
                if "data" in link:
                    data = link["data"]
                    for state_id in link:
                        if not isinstance(state_id, int):
                            continue
                        s = state[state_id]
                        for path in link[state_id]:
                            key = path["key"]
                            ref = rpath.path_ref(s, key)
                            if isinstance(ref, list):
                                # TODO check key ordering
                                if len(ref) == key[-1]:
                                    ref.append(data)
                                else:
                                    ref[key[-1]] = data
                            else:
                                ref[key[-1]] = data

            state = low.low(state)
            states.append(state)
    states = [high.high(s) for s in states]
    return states



import argparse
import shlex

from idem_data_insights.idem_data_insights import analyze
from idem_data_insights.idem_data_insights import group

WIZ_VIEWER = argparse.ArgumentParser(
    description="idem_data_insights - Wiz: A wizard that helps breaking up States."
)

WIZ_VIEWER.add_argument(
    "-q", "--quit", default=False, action="store_true", help="quit wizard"
)

WIZ_VIEWER.add_argument(
    "-s",
    "--select",
    nargs="*",
    help="select states to moodily",
    type=int,
    required=False,
)

WIZ_VIEWER.add_argument(
    "-p", "--picked", default=False, action="store_true", help="show picked salt states"
)

WIZ_VIEWER.add_argument(
    "-c", "--count", default=False, action="store_true", help="show state count"
)

WIZ_VIEWER.add_argument(
    "-m", "--mutable", default=False, action="store_true", help="show mutable"
)

WIZ_VIEWER.add_argument(
    "-v", "--variation", default=False, action="store_true", help="show variation"
)

WIZ_VIEWER.add_argument(
    "-g", "--group", default=False, action="store_true", help="group on states"
)

WIZ_VIEWER.add_argument(
    "-n", "--null_split", default=False, action="store_true", help="split all states"
)

WIZ_VIEWER.add_argument("-w", "--path_split", default=None, help="split on path value")

WIZ_VIEWER.add_argument(
    "-o", "--other", nargs="*", help="select 3rd party state", type=str, required=False
)


def _get_args(command=None):
    if command is None:
        command = input(">>> ")
    command = shlex.split(command)
    try:
        return WIZ_VIEWER.parse_args(command)
    except SystemExit:
        pass


def _get_selected(data, selected, inverse=False):
    if not selected:
        if inverse:
            return []
        return data
    selected_data = []
    for spot, states in data:
        if spot not in selected and inverse:
            selected_data.append(states)
        else:
            selected_data.append(states)
    return selected_data


def wiz_runner(data, selected, command, other):
    command = _get_args(command)
    if command is not None:
        if command.quit:
            return False
        if command.select is not None:
            if not command.select:
                selected.clear()
            else:
                for i in command.select:
                    if i < 0 or i >= len(data):
                        print(f"{repr(i)} is not in range of a salt state.")
                        continue
                    selected.add(i)
        if command.picked:
            if not len(selected):
                print(set(range(len(data))))
            else:
                print(selected)
        if command.count:
            analyze.state_count_pp(analyze.state_count(_get_selected(data, selected)))
        if command.mutable:
            analyze.mutable_data_pp(analyze.mutable_data(_get_selected(data, selected)))
        if command.variation:
            analyze.variation_sort_pp(
                analyze.variation_sort(_get_selected(data, selected))
            )
        if command.group:
            groups = group.split(_get_selected(data, selected), group.state_fun_comp)
            non_groups = _get_selected(data, selected, True)
            data.clear()
            data.extend(groups)
            data.extend(non_groups)
            selected.clear()
        if command.null_split:
            groups = group.split(_get_selected(data, selected), group.null_comp)
            non_groups = _get_selected(data, selected, True)
            data.clear()
            data.extend(groups)
            data.extend(non_groups)
            selected.clear()
        if command.path_split is not None:
            # TODO support path indexing of arrays
            groups = group.split(
                _get_selected(data, selected),
                group.build_path_comp(tuple(command.path_split.split("."))),
            )
            non_groups = _get_selected(data, selected, True)
            data.clear()
            data.extend(groups)
            data.extend(non_groups)
            selected.clear()
        if command.other is not None:
            # TODO handle errors on bad keys or out of index
            other[command.other[0]](command.other(data, selected, *command.other[1:]))
    return True


def wiz(data, other=None):
    if other is None:
        other = {}
    selected = set()
    _get_args("-h")
    while wiz_runner(data, selected, input(">>> "), other):
        pass
    return data



from idem_data_insights.idem_data_insights.low import INSIGHTS_LOW
from idem_data_insights.idem_data_insights.low import PATHS
from idem_data_insights.idem_data_insights.low import PATHS_INVERSE


STATES = "STATES"
STATE_COUNT = "STATE_COUNT"
PATH_STATES = "PATH_STATES"
PATHS_DATA = "PATHS_DATA"
DATA_PATHS = "DATA_PATHS"
OPTIONAL = "OPTIONAL"


PATH_VARIATIONS = "PATH_VARIATIONS"
COUNT = "COUNT"
UNIQUE = "UNIQUE"
AVERAGE = "AVERAGE"
VARIATIONS = "VARIATIONS"

MUTABLE = "MUTABLE"
IMMUTABLE = "IMMUTABLE"

PARENTS = "PARENTS"
CHILDREN = "CHILDREN"


def high(sls_data):
    high_data = {
        STATES: sls_data,
        STATE_COUNT: {},
        PATH_STATES: {},
        PATHS_DATA: {},
        DATA_PATHS: {},
        OPTIONAL: {},
        MUTABLE: set(),
        IMMUTABLE: set(),
        PATH_VARIATIONS: {},
        PARENTS: set(),
        CHILDREN: set(),
    }
    # STATE_COUNT
    for state in sls_data:
        fun_call = f"{state['state']}.{state['fun']}"
        if fun_call not in high_data[STATE_COUNT]:
            high_data[STATE_COUNT][fun_call] = 0
        high_data[STATE_COUNT][fun_call] += 1

    # PATHS_STATES
    for index, state in enumerate(sls_data):
        for path in state[INSIGHTS_LOW][PATHS]:
            if path not in high_data[PATH_STATES]:
                high_data[PATH_STATES][path] = set()
            high_data[PATH_STATES][path].add(index)

    # PATHS DATA
    for state in sls_data:
        for path, data in state[INSIGHTS_LOW][PATHS].items():
            if path not in high_data[PATHS_DATA]:
                high_data[PATHS_DATA][path] = set()
            high_data[PATHS_DATA][path].add(data)

    # DATA PATHS
    for index, state in enumerate(sls_data):
        for data in state[INSIGHTS_LOW][PATHS_INVERSE]:
            if data not in high_data[DATA_PATHS]:
                high_data[DATA_PATHS][data] = set()
            high_data[DATA_PATHS][data].add(index)

    # OPTIONAL
    for path, states in high_data[PATH_STATES].items():
        high_data[OPTIONAL][path] = len(states) != len(sls_data)

    # PATH_VARIATION
    for path in high_data[PATH_STATES]:
        state_set = high_data[PATH_STATES][path]
        variations = {COUNT: len(state_set), VARIATIONS: {}}
        high_data[PATH_VARIATIONS][path] = variations
        for state_id in state_set:
            state_data = sls_data[state_id][INSIGHTS_LOW][PATHS][path]
            if state_data not in variations[VARIATIONS]:
                variations[VARIATIONS][state_data] = set()
            variations[VARIATIONS][state_data].add(state_id)
        variations[UNIQUE] = (len(variations[VARIATIONS]) / variations[COUNT]) * 100
        variations[AVERAGE] = sum(
            len(variations[VARIATIONS][v]) for v in variations[VARIATIONS]
        ) / len(variations[VARIATIONS])

    # MUTABLE
    for path, variations in high_data[PATH_VARIATIONS].items():
        if len(variations[VARIATIONS]) != 1 or high_data[OPTIONAL][path]:
            high_data[MUTABLE].add(path)

    # IMMUTABLE
    for path in high_data[PATH_STATES]:
        if path not in high_data[MUTABLE]:
            high_data[IMMUTABLE].add(path)

    # PARENTS
    for path in high_data[PATH_STATES]:
        if len(path) <= 1:
            continue
        high_data[PARENTS].add(path[:-1])

    # CHILDREN
    for path in high_data[PATH_STATES]:
        if path not in high_data[PARENTS]:
            high_data[CHILDREN].add(path)
    return high_data
